import click

from reliesl.cli.bump import bump
from reliesl.cli.release import release


@click.group()
def reliesl():
    pass


reliesl.add_command(bump)
reliesl.add_command(release)


def main():
    """Entry point to be used when building reliesl as a library."""
    reliesl()



import os
from pathlib import Path
from typing import Dict
from typing import List
from typing import Union
from urllib.parse import urljoin

import click
import requests
import semver

from reliesl.cli.bump import get_version_and_name
from reliesl.cli.bump import load_changelog

# Type Definitions
Header = Dict[str, str]
Asset = Dict[str, str]
Json = object


def get_project_data(project_api_url: str, auth: Header) -> Json:
    project_data = requests.get(project_api_url, headers=auth)
    project_data.raise_for_status()
    return project_data.json()


def upload_file(path: Path, project_api_url: str, auth: Header) -> Json:
    uploads_url = project_api_url + "/uploads"
    file = {"file": (path.name, path.open("rb"), "application/zip")}
    result = requests.post(uploads_url, files=file, headers=auth)
    result.raise_for_status()
    return result.json()


def release_version(
    semver: str,
    release_description: str,
    assets: Union[Asset, List[Asset]],
    project_api_url: str,
    auth: Header,
    ref: str = "master",
) -> Json:
    if not isinstance(assets, list):
        assets = [assets]
    release_url = project_api_url + "/releases"
    release_name = "v" + semver
    release_data = {
        "name": release_name,
        "tag_name": release_name,
        "ref": ref,
        "description": release_description,
        "assets": {"links": assets},
    }
    auth["Content-Type"] = "application/json"
    result = requests.post(release_url, json=release_data, headers=auth)
    result.raise_for_status()
    return result.json()


def upload_wheel(
    wheel_path: Path, project_api_url: str, project_url: str, auth: Header
) -> Asset:
    wheel_reply = upload_file(
        path=wheel_path, project_api_url=project_api_url, auth=auth
    )
    wheel_asset = {"name": wheel_reply["alt"], "url": project_url + wheel_reply["url"]}
    return wheel_asset


def find_wheel(project_name: str, version: str) -> Path:
    click.echo(
        f"Trying to find wheel... - {project_name}-{semver.finalize_version(version)}*.whl"
    )
    wheel_path = list(
        Path("dist/").glob(f"{project_name}-{semver.finalize_version(version)}*.whl")
    )
    if len(wheel_path) != 1:
        raise SystemExit(
            f"Wheel could either not be found or is ambiguous: {wheel_path}"
        )
    wheel_path = wheel_path[0]
    return wheel_path


def get_project_url(project_api_url: str, auth: Header) -> str:
    try:
        project_data = get_project_data(project_api_url=project_api_url, auth=auth)
    except requests.exceptions.HTTPError:
        raise SystemExit("Could not connect to gitlab api. Is the given token correct?")
    project_url = project_data["web_url"]
    return project_url


def get_commit(branch: str, project_api_url: str, auth: Header) -> str:
    result = requests.get(
        project_api_url + f"/repository/branches/{branch}", headers=auth
    )
    result.raise_for_status()
    return result.json()["commit"]["short_id"]


def check_for_version_conflict(version: str, project_api_url: str, auth: Header):
    result = requests.get(project_api_url + f"/releases", headers=auth)
    result.raise_for_status()
    result = result.json()
    if result:
        current_version = result[0]["tag_name"]
        if semver.compare(version, current_version[1:]) != 1:
            raise SystemExit(
                f"Release version v{version} is not newer than current version {current_version}!"
            )


def _get_release_notes() -> str:
    changelog = load_changelog()  # Only interested in newest release notes
    newest_release = 0
    try:  # Check if Release 0 is not an unreleased version
        semver.parse(changelog[newest_release]["version"])
    except ValueError:
        newest_release = 1
    patch = changelog[newest_release]
    return patch["release_notes"]


def _parse_project_url(ctx, param, value: str) -> str:
    if not value:
        value = "gitlab.com"
    if not value.endswith("/"):
        value += "/"
    if not value.startswith("http"):
        value = "https://" + value
    return value


@click.command(help="Release a new version of your project on GitLab.")
@click.option(
    "--project-dir",
    envvar="CI_PROJECT_DIR",
    type=click.Path(exists=True),
    help="Root directory of the project you want to release.",
)
@click.option(
    "--project-id",
    envvar="CI_PROJECT_ID",
    help="Unique GitLab ID for projects, available in Project Settings > General.",
)
@click.option(
    "--project-url",
    callback=_parse_project_url,
    envvar="CI_PROJECT_URL",
    help="URL of GitLab server.",
)
@click.option(
    "--pypi-release", type=click.Choice(["pypi", "testpypi"]), help="Release to PyPi"
)
def release(project_dir: str, project_id: str, project_url: str, pypi_release: str):
    project_dir = Path(project_dir)
    branch = os.environ.get("CI_COMMIT_REF_NAME")
    private_token = os.environ.get("PRIVATE_TOKEN")
    if not private_token:
        raise SystemExit("No private token given. Set $PRIVATE_TOKEN")
    auth = {"PRIVATE-TOKEN": private_token}
    if not branch:
        raise SystemExit("No branch specified. Set $CI_COMMIT_REF_NAME")
    if pypi_release:
        env_username = pypi_release.upper() + "_USERNAME"
        env_password = pypi_release.upper() + "_PASSWORD"
        username = os.environ.get(env_username)
        password = os.environ.get(env_password)
        if not (username and password):
            raise SystemExit(
                f"Try to publish to {pypi_release}, but environment variable {env_username} or {env_password} not set!"
            )
    # Change to the projects directory in a revertible way
    prevdir = Path.cwd()
    os.chdir(project_dir.resolve())
    try:
        version, name = get_version_and_name()
        project_api_url = urljoin(project_url, f"/api/v4/projects/{project_id}")
        project_url = get_project_url(project_api_url, auth)

        check_for_version_conflict(version, project_api_url, auth)

        commit = get_commit(branch, project_api_url, auth)
        click.echo(f'Release v{version} on branch "{branch}", commit "{commit}"')
        os.system(f"poetry build")
        wheel_path = find_wheel(name, version)

        wheel_asset = upload_wheel(
            wheel_path=wheel_path,
            project_api_url=project_api_url,
            auth=auth,
            project_url=project_url,
        )
        release_version(
            semver=version,
            release_description=_get_release_notes(),
            assets=wheel_asset,
            project_api_url=project_api_url,
            ref=branch,
            auth=auth,
        )
        if pypi_release:
            if pypi_release == "testpypi":
                os.system(
                    f"poetry publish -r {pypi_release} -u {username} -p {password}"
                )
            else:
                os.system(f"poetry publish -u {username} -p {password}")
    finally:
        os.chdir(prevdir)


if __name__ == "__main__":  # Helpful for debugging
    release(["--pypi-release", "testpypi"])



import datetime
import os
import re
from pathlib import Path
from typing import Dict
from typing import List
from typing import Tuple

import click
import semver
import toml

CHANGELOG = Path("CHANGELOG.md")
PYPROJECT = Path("pyproject.toml")
PYPROJECT_BUMP_CONFIG = (PYPROJECT, '(version = ")(.*)(")')
EMPTY_PATCH = {
    "version": "UNRELEASED",
    "date": "YYYY-MM-DD\n",
    "release_notes": "\n### Added\n### Fixed\n### Changed\n\n",
}

RELEASE_CHOICES = {
    "1": {
        "description": "MAJOR version when you make incompatible API changes",
        "function": semver.bump_major,
    },
    "2": {
        "description": "MINOR version when you add functionality in a backwards compatible manner",
        "function": semver.bump_minor,
    },
    "3": {
        "description": "PATCH version when you make backwards compatible bug fixes",
        "function": semver.bump_patch,
    },
}
PRERELEASE_CHOICES = {
    "4": {
        "description": "RELEASE version when you want to release this version",
        "function": semver.finalize_version,
    },
    "5": {
        "description": "PRERELEASE version when you want to add a new prerelease",
        "function": semver.bump_prerelease,
    },
}


# Type Definitions
Releasenotes = Dict[str, str]
Changelog = List[Releasenotes]


def _get_date() -> str:
    return datetime.date.today().strftime("%Y-%m-%d")


def load_changelog(changelog_file: Path = CHANGELOG) -> Changelog:
    """

    Args:
        changelog_file: path to changelog in keepachangelog.com format

    Returns: List of releasenote dicts with the following fields:
        - version: Version number of the patch
        - date: Date of the patch
        - release_notes: Markdown formatted release_notes

    """
    with changelog_file.open() as file:
        changelog = []
        current_version = None
        for line in file:
            if line.startswith("## "):
                # Guarantees that only valid release_notes are added
                if current_version:
                    changelog += [current_version]
                parts = line.split(" ")
                patch_version = parts[1]
                try:  # Additional check in case no date is given and as such can not be parsed
                    patch_date = parts[3]
                except IndexError:
                    patch_date = "YYYY-MM-DD"
                current_version = {
                    "version": patch_version,
                    "date": patch_date,
                    "release_notes": "",
                }
            else:
                if current_version:
                    current_version["release_notes"] += line
    changelog += [current_version]
    return changelog


def _format_release_notes(release_notes: Releasenotes) -> str:
    return f"## {release_notes['version']} - {release_notes['date']}{release_notes['release_notes']}"


def save_changelog(changelog: Changelog, filename: Path):
    """

    Args:
        changelog: Changelog in the same format as returned by load_changelog()
        filename: filepath under which the changelog will be saved

    """
    click.echo("Saving changelog...")
    tmpfile = filename.with_suffix(".tmp")
    with tmpfile.open("w+") as file:
        for release in changelog:
            file.write(_format_release_notes(release))
    tmpfile.rename(filename)


def change_version_in_files(
    additional_file_edits: List[Tuple[Path, str]], version: str, dry_run: bool = False
):
    """

    Args:
        additional_file_edits: List of tuples containing a filepath and a regexpr
        version: Changed Version
        dry_run: Flag to check if no edits should be made. Helpful when debugging
            the regular expression.

    """
    for file, regexp in additional_file_edits:
        # Use tmpfile in case of crash, will replace original later
        # Careful when using, you need to guarantee that the whole line gets matched
        click.echo(f"Changes in {file} :")
        tmp_file = file.with_suffix(".tmp")
        with file.open("r") as realfile, tmp_file.open("w+") as tmpfile:
            if dry_run:
                click.echo(f"{regexp} finds: ")
            regexp = re.compile(regexp)
            for line in realfile:
                re.compile(regexp)
                line_match = regexp.match(line)
                if line_match:
                    click.echo(f"\t{line[:-1]} -> ", nl=False)
                    matches = list(line_match.groups())
                    matches[1] = version
                    line = "".join(matches) + "\n"
                    click.echo(line)
                tmpfile.write(line)
        if dry_run:
            tmp_file.unlink()
        else:
            tmp_file.rename(file)


def ask_patchtype(project_current_version: str):
    click.echo(f"The current version is: {project_current_version}")
    click.echo("What kind of update do you want to make?")
    if semver.parse(project_current_version)["prerelease"]:
        RELEASE_CHOICES.update(PRERELEASE_CHOICES)
    for key, value in RELEASE_CHOICES.items():
        click.echo(
            f"({key}) {value['function'](project_current_version)}\t - {value['description']}"
        )
    click.echo("For more information visit https://semver.org/")
    patch_type = click.prompt("> ", type=click.Choice(RELEASE_CHOICES.keys()))
    patch_type = RELEASE_CHOICES[patch_type]
    return patch_type["function"](project_current_version)


def _show_release_notes(release_notes: Releasenotes):
    click.echo(_format_release_notes(release_notes))
    click.echo("-----")
    click.confirm(
        "This is how the new release_notes will look like. Do you want to continue?",
        abort=True,
    )


def get_version_and_name(pyproject_file: Path = PYPROJECT):
    with pyproject_file.open() as file:
        pyproject = toml.load(file)
    project_pyproject = pyproject["tool"]["poetry"]
    project_current_version = project_pyproject["version"]
    project_name = project_pyproject["name"]
    return project_current_version, project_name


def _parse_file_edits(ctx, params, value):
    return [(Path(file_regex[0]), file_regex[1]) for file_regex in value]


@click.command(help="Bump the version number for your project.")
@click.option(
    "--project-dir",
    default=Path.cwd(),
    callback=lambda ctx, params, value: Path(value),
    help="Directory in which to look for the CHANGELOG.md and pyproject.toml",
)
@click.option(
    "--additional-file-edits",
    "-f",
    nargs=2,
    multiple=True,
    callback=_parse_file_edits,
    help="Specify <file> <regex> for additional files and a corresponding regex where "
    "the version number needs to be bumped.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Do a dry run to check if file edits work correctly without changing any files.",
)
@click.option(
    "--prerelease",
    is_flag=True,
    default=False,
    help="Change new version to a prerelease.",
)
def bump(
    project_dir: Path, additional_file_edits: List[str], dry_run: bool, prerelease: bool
):
    additional_file_edits.append(PYPROJECT_BUMP_CONFIG)
    # Change to the projects directory in a revertible way
    prevdir = Path.cwd()
    os.chdir(project_dir.resolve())
    try:
        if dry_run:
            click.echo("It's a dry run.")
            click.echo("The following file edits would be made:")
            change_version_in_files(
                additional_file_edits, version="<NEWVER>", dry_run=True
            )
            click.echo(
                "\nP.S. Be careful to capture the whole line, not captured parts will be deleted!"
            )
            exit(0)

        project_current_version, project_name = get_version_and_name()
        new_version = ask_patchtype(project_current_version)
        if prerelease:
            new_version = semver.bump_prerelease(new_version)
        changelog = load_changelog(CHANGELOG)

        if changelog[0]["version"] != "UNRELEASED":
            raise SystemExit(
                "No new release_notes detected. (Start with '## UNRELEASED')"
            )

        changelog[0]["version"] = new_version
        changelog[0]["date"] = _get_date() + "\n"
        _show_release_notes(changelog[0])

        # Start of all permanent changes
        changelog.insert(0, EMPTY_PATCH)
        save_changelog(changelog, CHANGELOG)

        # Bump in all other files
        change_version_in_files(additional_file_edits, version=new_version)
    finally:
        os.chdir(prevdir)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['reliesl', 'reliesl.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'requests>=2.22,<3.0',
 'semver>=2.8,<3.0',
 'toml>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['reliesl = reliesl.cli:main']}

setup_kwargs = {
    'name': 'reliesl',
    'version': '0.1.1',
    'description': 'Reliesl is a tool for automatic release managment using the Gitlab CI.',
    'long_description': '<p align="center">\n<a href="https://github.com/ambv/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\n# reliesl\nreliesl helps you automate your releases using `poetry` and GitLab CI.\n\n<div align="center">\n    <img src="assets/liesl.jpg" alt="Liesl" width=200px align="middle">\n</div>\n\n"You\'ve had a long day at work. Bugs have been squashed, features wered added and breaking changes documented. You\'ve poured your heart and soul into this piece of software and the last thing seperating you from popping a bottle of champagne is having to release the newest version of your project on GitLab. Thank God there\'s reliesl to release your new software on GitLab (and maybe even open that bottle of champagne)."\n\\- reliesl\'s secret admirer\n\n## Use `reliesl`\n`reliesl` grew up in a small village and thus is pretty opinionated. It assumes you are using `poetry` and keeping a `CHANGELOG.md`.\n`reliesl` provides two different commands to make your release process as smooth as possible:\n\n### reliesl release\nUsually added to your existing CI pipeline to manually trigger a new GitLab release based on the releasenotes given in CHANGELOG.md. Also adds the .whl of the package to the Asset section.\n\nIf you want to automatically release to PyPi (or TestPyPi if you are feeling less adventurous) you can use the `--pypi-release <pypi|testpypi>` option like so:\n```bash\nreliesl release --pypi-release <value>\n```\nFor authentication this requires the environement variables  `PYPI_USERNAME`  and `PYPI_PASSWORD` or `TESTPYPI_USERNAME` and `TESTPYPI_PASSWORD` to be set.\n\n### reliesl bump\nUsually run locally to take care of all cumbersome version edits in various files. \nIt will ask you whether you want to do a pre-release, bug fix, minor or major change, before automatically bumping your `poetry` version and parsing the `CHANGELOG.md` to update the `UNRELEASED` section with the right version number and date.\n\nUse the `--dry-run` flag to see what would happen without changing any files.\n\nAdditional files where version changes are needed can be configured via `--additional-file-edits <file> <regex>`, like so: \n\n```\nreliesl bump --additional-file-edits "file.txt" \'(semver = )(.*)()\' "file2.txt" \'(version: )(.*)()\'\n```\n\nThe second capture group must be the semantic version you want to change.\nFor more information on python regular expressions look [here](https://docs.python.org/3.7/howto/regex.html).\n\n## Install `reliesl`\n\n### Local\nSimply run\n```bash\npoetry add --dev reliesl\n```\nor \n```bash\npip install reliesl\n```\n\n### GitLab-CI\nTo your .gitlab-ci.yml you need to add:\n\n - A new `- deploy` stage if you do not already have one (The release process should be the last stage)\n\n - A new job:\n```\nrelease:\n  stage: deploy\n  when: manual\n  script:\n    - pip install reliesl\n    - reliesl release\n```\n\n\n### Environment Variables\nThe release script needs access to the GitLab API to create a new release. This requires a new environment variable(Under Settings > CI / CD > Variables):\n * Key: `PRIVATE_TOKEN`\n * Value: a valid private token that gives access to api calls for your project(Profile > Settings > Access Tokens)\n\n\n## Development\n\n### Pre-commit\nPlease make sure to have the pre-commit hooks installed.\nInstall [pre-commit](https://pre-commit.com/) and then run `pre-commit install` to register the hooks with git.\n\n### Makefile\nWe use [make](https://www.gnu.org/software/make/) to streamline our development workflow.\nRun `make help` to see all available commands.\n\n<!-- START makefile-doc -->\n```\n$ make help \nhelp                 Show this help message\ncheck                Run all static checks (like pre-commit hooks)\ntest                 Run all tests\ndev-install          Install all the packages in the local python environment for development\nbump                 Bump reliesl version \n```\n<!-- END makefile-doc -->\n\n## Acknowledgments\n\nThe picture of Liesl was stolen from here: https://ladiesoftheatre.tumblr.com/post/155234955388/you-may-think-this-kind-of-adventure-never-may\n',
    'author': 'Benedikt Wiberg',
    'author_email': 'benedikt@luminovo.ai',
    'url': 'https://gitlab.com/luminovo/public/reliesl',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
